;; $Id: convert.lisp,v 1.3 2007/07/26 08:58:08 willijar Exp $
;;
;; Copyright (C) 2006 Dr. John A.R. Williams

;; Author: Dr. John A.R. Williams <J.A.R.Williams@jarw.org.uk>
;; Keywords: d

;; This file is part of the clews.media library

;; This is free software released under the GNU General Public License (GPL)
;; See <http://www.gnu.org/copyleft/gpl.html>

;;; Commentary:

;;
;;

;;; Code:

(in-package :jarw.media)

(defvar *media-conversions* (make-hash-table :test #'equalp))

(defvar *media-formats*  (make-hash-table :test #'equalp))

(defvar *file-suffixes* (make-hash-table :test #'equalp))

(defun file-suffix-content-type(suffix)
  (when (pathnamep suffix) (setf suffix (pathname-type suffix)))
  (gethash suffix *content-types* "application/octet-stream"))

(defun content-type-file-suffix(content-type)
  (or (gethash content-type *file-suffixes*)
      (if (is-prefix-p "text" content-type :test #'char-equal)
          "txt" "bin")))

(defun register-media-format(content-type element-type external-format
                             &optional file-suffix)
  "Register the internal element type and external file format for a
specific media type"
  (setf (gethash content-type *media-formats*)
        (list element-type external-format))
  (when file-suffix
    (setf (gethash content-type *file-suffixes*) file-suffix)
    (setf (gethash file-suffix *content-types*) content-type)))

(defun content-type-element-type(content-type)
  "Return the element type"
  (let ((entry (gethash content-type *media-formats*)))
    (cond
      (entry (first entry))
      ((is-prefix-p "text" content-type :test #'char-equal)
       'character)
      ('unsigned-byte))))

(defun content-type-external-format(content-type)
  "Return the external file format for a given content type"
  (let ((entry (gethash content-type *media-formats*)))
    (cond
      (entry (second entry))
      (:default))))

(defun read-media(path &key (content-type (file-suffix-content-type path)))
  (let ((type (content-type-element-type content-type)))
    (with-open-file
        (is path
            :direction :input
            :external-format
            (content-type-external-format content-type)
            :element-type type)
      (let ((result (make-array (file-length is) :element-type type)))
        (read-sequence result is)
        result))))

(defun write-media(media path
                   &key (content-type (file-suffix-content-type path)))
  (let ((type (content-type-element-type content-type)))
    (with-open-file
        (os path
            :direction :output
            :if-does-not-exist :create
            :if-exists :supersede
            :external-format (content-type-external-format content-type)
            :element-type type)
      (write-sequence media os))))

(defmacro def-media-conversion(((informat outformat)
                                &rest args)
                               &body body)
  "Define a new media conversion routine. invar and outvar are the
variable names for the input and output pathnames respectively. Should
return the destination outpath, or a list of steps to carry our
conversion. Each step is a list of the content-type and the arguments
to be used."
  `(setf (gethash (list ,informat ,outformat) *media-conversions*)
        #'(lambda(,@(or args '(in out)))
             ,@(when (not args) '((declare (ignore in out))))
             ,@body)))

(defgeneric convert-media(intype outtype source destination
                                 &key &allow-other-keys)
  (:documentation "Convert media from source (content type intype)
writing to destination (content type outtype) using keyword
args. Returns destination"))

(defmethod convert-media(informat outformat
                         (inpath pathname) (outpath pathname)
                         &key args)
  "Convert media from a source file to a destination file"
  (when (and (not args) (equalp informat outformat))
    (unless (equalp inpath outpath)
      (copy-file inpath outpath))
    (return-from convert-media outpath))
  (let ((conversion (gethash (list informat outformat) *media-conversions*)))
    (unless conversion (error "No conversion from ~S ~A to ~S ~A found"
                              informat inpath outformat outpath))
    (let ((result (apply conversion `(,inpath ,outpath ,@args))))
      (if (listp result)
          (let ((informat informat)
                (inpath inpath)
                (temp-paths nil))
            (unwind-protect
                 (progn
                   (dolist(step result)
                     (let* ((outformat (car step))
                           (args (rest step))
                           (outpath (jarw.io::create-temp-pathname
                                     :type (content-type-file-suffix outformat))))
                       (apply #'convert-media
                              `(,informat ,outformat ,inpath ,outpath :args ,args))
                       (push outpath temp-paths)
                       (setf informat outformat
                             inpath outpath)))
                   (copy-file inpath outpath :if-new-file-exists :overwrite))
              (dolist(temp-path temp-paths)
                (jarw.io:delete-file-all-types temp-path))))
          result))))

(defmethod convert-media(informat outformat (inpath string) outpath
                         &key args)
  (with-temp-pathname(src :type (content-type-file-suffix informat))
    (with-open-file(os src :direction :output
                       :if-does-not-exist :create
                       :if-exists :overwrite)
      (write-string inpath os))
    (let ((dest (apply
                 #'convert-media
                 `(,informat ,outformat ,src
                   ,(if (typep outpath 'pathname)
                        outpath
                        (let ((dest
                               (create-temp-pathname
                                :type (content-type-file-suffix outformat))))
                          (delete-file dest)
                          dest))
                   ,@args))))
      (if (typep outpath 'pathname)
          dest
          (prog1
              (read-media dest :content-type outformat)
            (delete-file-all-types dest))))))

;;; Now some conversions

(defparameter *latex-prefix*
"\\documentclass[12pt]{article}
\\usepackage{graphicx}
\\usepackage{color}
\\usepackage{amsmath}
\\pagestyle{empty}
\\newcommand{\\mat}[1]{{\\bf #1}}
\\newcommand{\\erf}{\\mbox{erf}}
\\newcommand{\\erfc}{\\mbox{erfc}}
\\newcommand{\\sinc}{\\mbox{sinc}}
\\begin{document}
"
"Initial component of latex document")

(defparameter *latex-suffix*
"
\\end{document}"
  "Final part of latex document")

(defparameter *screen-dpi* 75 "Dots per inch on typical display")

(defparameter *length-units*
  '((:in . 1)
    (:cm . 254/100)
    (:em . 72/10)
    (:ex . 10)
    (:px . 75) ;; assume 75 dpi
    (:% . 75/8) ;;  100% is 800 pixels
    (:pt . 72)
    (:pc . 12/72)
    (:mm . 254/10))
  "Conversion from various units to inches")

(defun length-unit(unit)
  (or (cdr (assoc unit *length-units*))
      (jarw.parse:invalid-input unit "Acceptable units are ~S"
             (mapcar #'car *length-units*))))

(defun convert-length-unit(size unit)
  (unless (consp size) (setf size (cons size :px)))
  (cons
   (* (/ (car size) (length-unit (cdr size))) (length-unit unit))
   unit))

(defun latex-length(size)
  (case (cdr size)
    (:% (format nil "~A\textwidth" (/ (car size) 100.0)))
    (t  (format nil "~A~A" (car size) (cdr size)))))

(defun image-out-cmd(type dest &key (scale 1.0) width height (margin 1))
  "construct string of commands for shell to create output to given file"
  (when width (setf width (floor (car (convert-length-unit width :px)))))
  (when height (setf height (floor (car (convert-length-unit height :px)))))
  (when (or width height) (setf scale nil))
  (format nil
          "pnmcrop -white | pnmmargin -white ~D | pnmscale ~@[~A~] ~@[-width=~D~] ~@[-height=~D~] | ~A > \"~A\""
          margin
          scale
          width
          height
          (vcase (type  :test #'string-equal)
             ("image/gif"  "ppmquant 256 | ppmtogif -interlace -")
             ("image/png"  "pnmtopng -")
             ("image/jpeg" "pnmtojpeg -")
             ("image/x-eps"  "pnmtops -noturn -nosetpage -equalpixels -rle -")
             ("application/postscript"
              "pnmtops -noturn -nosetpage -equalpixels -rle -"))
          dest))

(defun image-to-pnm-cmd(type src &key size (dpi 300))
  "construct string of commands for shell to create output to given file"
  (if (member type '("image/x-eps" "application/postscript" "application/x-pdf"
                     "application/pdf") :test #'string-equal)
      (format nil "gs -sDEVICE=ppmraw -sPAPERSIZE=a3 -sOutputFile=- ~@[-g~Dx~:*~D~] -r~D -q -dNOPAUSE \"~A\" < /dev/null | " size dpi src)
      (format nil "~A < \"~A\" | "
              (vcase (type  :test #'string-equal)
                     ("image/gif" "giftopnm")
                     ("image/png""pngtopnm")
                     ("image/jpeg" "jpegtopnm")
                     ("image/x-eps"  "pstopnm -stdout")
                     ("application/postscript" "pstopnm -stdout")
                     ("application/pdf" "pstopnm -stdout")
                     ("application/x-pdf" "pstopnm -stdout"))
              src)))

(defun convert-image(informat outformat src dest &key (scale 1.0)
                     width height size (dpi 300)
                     (margin 1))
  (run-shell
   (concatenate 'string
                (image-to-pnm-cmd informat src :size size :dpi dpi)
                (image-out-cmd outformat dest :scale scale :width width
                               :height height :margin margin)))
  dest)

(eval-when(:load-toplevel :execute)
  ;; set up file suffixes from content types
  (maphash
   #'(lambda(k v) (setf (gethash v *file-suffixes*) k))
   *content-types*)
  ;; register some local formats and preferred defaults
  (register-media-format "image/x-eps" '(unsigned-byte 8) :default "eps")
  (register-media-format "image/jpeg" '(unsigned-byte 8) :default "jpeg")
  (register-media-format "text/x-rst" 'character :utf-8 "rst")
  (register-media-format "application/x-tex" 'base-char :ascii "tex")
  (register-media-format "application/x-latex" 'base-char :ascii "tex")
  (register-media-format "text/x-eqn" 'base-char :ascii "eqn")
  (register-media-format "text/x-mpost" 'base-char :ascii "mp")
  ;; register conversions between image formats
  (let ((types '("image/gif" "image/png" "image/jpeg" "image/x-eps"
                   "application/postscript" "application/pdf"
                    "application/x-pdf")))
  (dolist(informat types)
    (dolist(outformat types)
      (setf (gethash (list informat outformat) *media-conversions*)
            #'(lambda(inpath outpath &key  (dpi 300)
                      size
                      (scale 1.0) width height (margin 1))
                (convert-image informat outformat inpath outpath
                               :width width :height height :dpi dpi :size size
                               :scale scale :margin margin)))))))

(def-media-conversion(("application/x-latex" "application/x-dvi")
                      src dest)
  (with-temp-pathname(tmp :type "tex")
    (copy-file src tmp :if-new-file-exists :overwrite)
    (run-shell
     (format
      nil
      "cd ~S; latex -interaction=nonstopmode \"~A\" ; latex -interaction=nonstopmode \"~A\""
      (directory-namestring tmp)
      tmp tmp))
    (copy-file (merge-pathnames (make-pathname :type "dvi") tmp)  dest
               :if-new-file-exists :overwrite)))

(def-media-conversion(("text/x-eqn" "image/x-eps")
                      src dest  &key (prefix *latex-prefix*)
                      (suffix *latex-suffix*))
  (with-temp-pathname(tmp :type "tex")
    (with-open-file(os tmp :direction :output :if-exists :overwrite)
      (write-sequence prefix os)
      (with-open-file(is src :direction :input :if-does-not-exist :error)
        (copy-stream is os))
      (write-sequence suffix os))
    (run-shell
     (format
      nil
      "cd ~S; latex -interaction=nonstopmode \"~A\"; dvips -f \"~A\" -o \"~A\""
      (directory-namestring tmp)
      tmp (merge-pathnames (make-pathname :type "dvi") tmp) dest))
    dest))

(def-media-conversion(("text/x-eqn" "application/x-latex") src dest
                      &key
                      (prefix *latex-prefix*)
                      (suffix *latex-suffix*))
  (with-open-file(os dest
                     :direction :output
                     :if-does-not-exist :create
                     :if-exists :overwrite)
    (write-sequence prefix os)
    (with-open-file(is src :direction :input :if-does-not-exist :error)
      (copy-stream is os))
    (write-sequence suffix os))
  dest)

(def-media-conversion(("application/x-dvi" "application/postscript") src dest)
    (run-shell (format  nil "dvips -f \"~A\" -o \"~A\"" src dest))
    dest)

(def-media-conversion(("application/x-dvi" "application/pdf") src dest)
    (run-shell (format  nil "dvipdf \"~A\"  \"~A\"" src dest))
    dest)

(def-media-conversion(("application/x-dvi" "application/x-pdf") src dest)
    (run-shell (format  nil "dvipdf \"~A\" \"~A\"" src dest))
    dest)

(def-media-conversion(("application/x-latex" "application/postscript"))
  '(("application/x-dvi") ("application/postscript")))

(def-media-conversion(("application/x-latex" "application/pdf"))
  '(("application/x-dvi") ("application/pdf")))

(def-media-conversion(("application/x-latex" "application/x-pdf"))
  '(("application/x-dvi") ("application/x-pdf")))

(defun eqn-to-image(fmt &key (dpi 100) (scale 1) (detail 2.5) (margin 1))
  (let ((work-dpi (* detail dpi))
        (size (round (* 12 detail dpi scale)))
        (reduct-ratio (/ 1.0 detail)))
    `(("image/x-eps")
      (,fmt :size ,size :dpi ,work-dpi :scale ,reduct-ratio
       :width nil :margin ,margin))))

(def-media-conversion(("text/x-eqn" "image/gif")
                      src dest
                      &key (dpi 100) (scale 1) (detail 2.5) (margin 1))
  (declare (ignore src dest))
  (eqn-to-image "image/gif" :detail detail :dpi dpi
                :scale scale :margin margin))

(def-media-conversion(("text/x-eqn" "image/png")
                      src dest
                      &key (dpi 100) (scale 1) (detail 2.5) (margin 1))
  (declare (ignore src dest))
  (eqn-to-image "image/png" :detail detail :dpi dpi
                :scale scale :margin margin))

(def-media-conversion(("text/x-eqn" "image/jpeg")
                      src dest
                      &key (dpi 100) (scale 1) (detail 2.5) (margin 1))
  (declare (ignore src dest))
  (eqn-to-image "image/jpeg" :detail detail :dpi dpi
                :scale scale :margin margin))



(defun get-bbox(src)
  (with-output-to-string(os)
    (run-shell "gs -sDEVICE=bbox -dNOPAUSE -dBATCH -q $0"
               :args (list (namestring src))
               :error os) ))

(defun set-bbox(src &optional
                (dest (merge-pathnames (make-pathname :type  "eps") src))
                (bbox (get-bbox src)))
  (flet ((begins-with(a b)
           (when (> (length a) (length b))
             (string-equal a b :end1 (length b)))))
    (with-open-file (in src :direction :input)
      (with-open-file(out dest :direction :output :if-exists :supersede)
        (let ((line (read-line in nil nil)))
          (if (search "EPSF-3.0" line)
              (write-line line out)
              (write-line "%!PS-Adobe-3.0 EPSF-3.0" out)))
        (write-line bbox out)
        (loop
         (let ((line (read-line in nil nil)))
           (unless line (return))
           (cond ((begins-with line "@PJL"))
                 ((begins-with line "%%BoundingBox:"))
                 (t (write-line line out)))))))))

(def-media-conversion(("application/postscript" "image/x-eps")
                      src dest &key (bbox (get-bbox src)))
  (set-bbox src dest bbox))

(def-media-conversion (("application/x-xfig"  "image/x-eps")
                       src dest)
  (run-shell (format nil "fig2dev -L eps \"~A\" \"~A\"" src dest))
  dest)

(def-media-conversion (("text/x-mpost" "image/x-eps")
                       src dest &key number)
  (with-temp-pathname(tmp :type "mp")
    (copy-file src tmp)
    (run-shell (format nil "mpost  \"~A\"" tmp))
    (let ((mps (merge-pathnames (make-pathname :type (format "~A" number))
                                tmp)))
      (run-shell "mps2eps \"~A\"; mv \"~A.eps\" \"~A\"" mps dest))
    dest))

(def-media-conversion (("text/x-mpost" "image/png")
                       src dest &key number)
  (declare (ignore src dest))
  `(("image/x-eps" :number ,number)
    ("image/png")))

(def-media-conversion (("text/x-mpost" "image/gif")
                       src dest &key number)
  (declare (ignore src dest))
  `(("image/x-eps" :number ,number)
    ("image/gif")))

(def-media-conversion (("image/x-mpost" "image/jpeg")
                       src dest &key number)
  (declare (ignore src dest))
  `(("image/x-eps" :number ,number)
    ("image/jpeg")))

