;; $Id $
;; Media server
;; Copyright (C) 2006 Dr. John A.R. Williams

;; Author: Dr. John A.R. Williams <J.A.R.Williams@jarw.org.uk>
;; Keywords: docutils, math, LaTeX


;; This is free software released under the GNU General Public License (GPL)
;; See <http://www.gnu.org/copyleft/gpl.html>

;;; Commentary:

;; The aim of this is to provide automatic conversion and caching of
;; media formats and the metadata associated with it.

;; Media sources (data or filepaths) are registered with a server
;; which then serves up the media in various formats upon request
;; and deals with caching etc.

;; Media type are identified by a typspec consisting of a MIME type and
;; a property list of arguments

;;; Code:

(in-package :jarw.media)

(defvar *media-server* nil
  "The global default figure server")
(defvar *media-base-url* "/media/")

;;; User API
(defgeneric register-media(source server &key content-type metadata)
  (:documentation "Register a media source for a media entity of given
type with the *media-server*. Returns the allocated unique key for
this media source. If already present the old name is returned. A list
of initial conversions may be provided to be carried out in the
background. Note if a source is registered more than once, only the
first registration sets content-type and metadata."))

(defgeneric media-variant(media-key server content-type &optional args)
  (:documentation "Return the variant key for a particular conversion of a
media based on content-type and args. Returns as a second value
whether the conversion is completed or not."))

(defgeneric media-variant-data(variant-key server)
  (:documentation "Return the media data for the given variant key
server. If a new conversion is needed will wait for the conversion to
complete before returning the data. Returns content-type as second
value."))

(defgeneric media-variant-path(variant-key server)
  (:documentation "Return the path to a media variant given by a
variant key and the content type."))

(defgeneric media-variant-url(variant-key server)
  (:documentation "Return the media data for the given content type
and key on server. If a new conversion is needed will wait for the
conversion to complete before returning the data."))

(defgeneric media-base-url(server)
  (:documentation "Return the base url for serving up media variants
from this server")
  (:method (server) (declare (ignore server)) *media-base-url*))

(defgeneric media-metadata(media-key server)
  (:documentation "Return the list of metadata stored for given
media. Each element is a cons of the metadata fieldname and its string
value "))

(defgeneric (setf media-metadata)(values media-key server)
  (:documentation "Set the metadata for this media - this must be a
list of records each of which is a cons of the metdata fieldname and
its string value"))

;;; Implementation
(defclass media-record(filestore-fields)
  ((media-source-path
    :initform nil
    :format (pathname :nil-allowed t)
    :field-name "Source-Path"
    :documentation "The source path to the media file")
   (content-type :type list
                 :initarg :content-type
                 :format rfc2822:content-type
                 :field-name "Content-Type")
   (variants :type list :format (read :multiplep t)
             :field-name "Variants"
             :initform nil
             :accessor variants
             :documentation "equal a-list of variants"))
  (:default-initargs :external-format :utf-8)
  (:metaclass  standard-parsed-class)
  (:documentation "The figure record"))

(defmethod print-object((record media-record) stream)
  (print-unreadable-object (record stream :type t :identity t)
    (princ (slot-value record 'media-source-path) stream)))

(defgeneric content-type(media-record)
  (:method((record media-record))
    (if (slot-boundp record 'content-type)
        (first (slot-value record 'content-type))
        (file-suffix-content-type (slot-value record 'media-source-path)))))

(defgeneric media-source-path(record)
  (:documentation "Return the path of the source of this media.")
  (:method ((record media-record))
    (or (slot-value record 'media-source-path)
        (merge-pathnames
         (make-pathname :type (content-type-file-suffix (content-type record)))
         (path record)))))

(defgeneric media-source(media-record)
  (:documentation "The media source as registered")
  (:method ((record media-record))
    (or (slot-value record 'media-source-path)
        (read-media (media-source-path record)
                    :content-type (content-type record)))))

(defun conversion-subkey(variant record)
  (position variant (variants record) :test #'equal))

(defgeneric media-key(media-record)
  (:method ((record media-record))
    (pathname-name (path record))))

(defgeneric variant-path(key record)
  (:documentation "Return the path and content-type of given variant of media")
  (:method((key (eql nil)) (record media-record))
    (values (media-source-path record) (content-type record)))
  (:method ((key integer) (record media-record))
    (let ((v (elt (variants record) key)))
      (values
       (merge-pathnames
        (make-pathname :name (format nil "~A-~A" (media-key record) key)
                       :type (content-type-file-suffix (first v)))
        (path record))
       (first v))))
  (:method ((variant list) (record media-record))
    (let ((p (position variant (variants record) :test #'equal)))
      (when p (variant-path p record)))))

(defmethod initialize-instance :after
    ((record media-record)
     &key source path metadata content-type &allow-other-keys)
  (setf (jarw.parse::unparsed-values record) metadata)
  (unless (listp content-type)
    (setf (slot-value record 'content-type) (list content-type)))
  (let ((content-type (content-type record)))
    (when source
      (etypecase source
        (pathname (setf (slot-value record 'media-source-path) source))
        (sequence
         (write-media
          source
          (merge-pathnames
           (make-pathname :type (content-type-file-suffix content-type))
           path)
          :content-type content-type))))))

(defgeneric metadata(media-record)
  (:method ((record media-record))
    (jarw.parse::unparsed-values record)))

(defgeneric (setf metadata) (values media-record)
  (:method(values (record media-record))
    (setf (jarw.parse::unparsed-values record) values)))

(defclass media-server()
  ((path :type pathname :initarg :path :reader path
         :documentation "Destination directory for storing media")
   (lock
    :initform (make-mutex) :reader lock
    :documentation "Master lock for this server")
   (pending-conversions
    :type list :initform nil :accessor pending-conversions
    :documentation "Queue of conversions to be performed")
   (pending-wait
    :initform (make-waitqueue) :reader pending-wait
    :documentation "Waitqueue for pending operations")
   (pending-lock
    :initform  (make-mutex) :reader pending-lock
    :documentation "Lock for pending operations")
   (pending-thread
    :reader pending-thread
    :documentation "Thread carrying out conversions in the background")
   (status
    :initform :initialising :accessor status
    :documentation "If set to true server is stopping")
   (media-sources :initform (make-hash-table :test #'equal)
                  :reader media-sources
                  :documentation "Mapping between sources and media keys")
   (media-records :reader media-records :type dictionary))
  (:documentation "The file server"))

(defmethod initialize-instance :after ((server media-server)
                                       &key (media-record-type 'media-record)
                                       &allow-other-keys)
  (assert (not (or (pathname-name (path server))
                   (pathname-type (path server))))
          ((slot-value server 'destination))
          "Figure Server destination must be a directory")
  (ensure-directories-exist (path server))
  (setf (slot-value server 'media-records)
        (make-instance 'filestore-class-dictionary
                       :path (merge-pathnames (make-pathname :name :wild)
                                              (path server))
                       :record-type media-record-type))
  (map-dictionary
   #'(lambda(key record)
       (setf (get-dictionary (media-source record) (media-sources server))
             key))
   (media-records server))
  (start-server server))


(defun stop-server(server &key (abort t))
  "Pending tasks will be aborted if abort is true (default)"
  (setf (status server)
        (if abort
            (ignore-errors
              (terminate-thread (pending-thread server))
              (dolist(pending (pending-conversions server))
                (setf (pending-status pending) :pending))
              (condition-broadcast (pending-wait server))
              :stopped)
            :stopping)))

(defun start-server(server)
  (stop-server server :abort t)
  (setf (slot-value server 'pending-wait)
        (make-waitqueue)
        (slot-value server 'pending-thread)
        (make-thread #'(lambda() (conversion-loop server))
                     :name "Media Conversion"))

  (setf (status server) :running))

(defun reload-server(server)
  (clr-cache (media-records server))
  (clr-dictionary (media-sources server))
  (map-dictionary
   #'(lambda(key record)
       (setf (get-dictionary (media-source record) (media-sources server))
             key))
   (media-records server)))

(defmethod register-media (source (server media-server) &key content-type
                           metadata)
  (unless content-type
    (setf content-type
          (etypecase source
            (pathname (file-suffix-content-type source))
            ((vector character )  "text/plain")
            ((vector (unsigned-byte 8))  "application/octet-stream"))))
  (or (gethash source (media-sources server))
      (let ((prefix (if (pathnamep source)
                        (pathname-name source)
                        (subseq content-type
                                (1+ (position #\/ content-type
                                              :from-end t))))))
        (setf (gethash source (media-sources server))
              (flet ((record-path(name)
                       (merge-pathnames (make-pathname :name name :type nil)
                                        (path server))))
                ;; we scan to find a unique name based upon name component
                ;; of source and then create a media-record
                (with-lock((lock server))
                  (do*((c 0 (1+ c))
                       (name (format nil "~A-~D" prefix c)
                             (format nil "~A-~D" prefix c))
                       (path (record-path name) (record-path name)))
                      ((not (probe-file path))
                       (progn
                         (let ((record
                                (make-instance 'media-record
                                               :path path
                                               :metadata metadata
                                               :source source
                                               :content-type (list content-type))))
                           (update-record-from-instance record)
                           (setf (get-dictionary (media-source record)
                                                 (media-sources server))
                                 name))
                         name)))))))))

(defstruct pending
  (priority 0 :type (integer -20 19))
  (media-record nil :type media-record)
  (conversion nil :type list)
  (status :pending :type (member :pending :converting :completed :error)))

(defun find-pending(conversion media-record server)
  (let ((source-path (media-source-path media-record)))
    (find-if
     #'(lambda(pending)
         (and
          (equal conversion (pending-conversion pending))
          (equal source-path
                 (media-source-path (pending-media-record pending)))))
     (pending-conversions server))))

(defun insert-in-order(item list test &key (key #'identity))
  (flet ((cmp(a b) (funcall test (funcall key a) (funcall key b))))
    (if (cmp item (first list))
        (cons item list)
        (loop :for a :on list :by #'cdr
              :do (when (or (not (cdr a)) (cmp item  (second a)))
                    (setf (cdr a) (cons item (cdr a)))
                    (return-from insert-in-order list))))))

(defun add-pending(conversion media-record server &optional (priority 0))
  "Adds a pending conversion to server, returning the pending record"
  (when (eql (status server) :stopping)
    (error "Attempt to add a task to a server which is stopping."))
  (with-lock ((pending-lock server))
    (let ((pending (find-pending  conversion media-record server)))
      (if pending
          (if (or (not priority)
                  (not (eql (pending-status pending) :pending))
                  (= priority (pending-priority pending)))
              (return-from add-pending pending)
              (progn
                (setf (pending-priority pending) priority)
                (setf (pending-conversions server)
                      (delete pending (pending-conversions server)))))
          (setf pending (make-pending :priority priority
                                      :media-record media-record
                                      :conversion conversion)))
      (setf (pending-conversions server)
            (if (pending-conversions server)
                (insert-in-order pending (pending-conversions server)
                                 #'< :key #'pending-priority)
                (list pending)))
      (condition-broadcast (pending-wait server))
      pending)))

(defun process-pending
    (server &optional
     (pending (find-if #'(lambda(pending)
                           (eql (pending-status pending)
                                :pending))
                       (pending-conversions server))))
  (with-lock ((pending-lock server))
    (assert (eql :pending (pending-status pending)))
    (setf (pending-status pending) :converting))
  (let* ((record (pending-media-record pending))
         (conversion (pending-conversion pending))
         (source (media-source-path record))
         (destination (variant-path conversion record )))
    (assert source (source) "Source file ~S does not exist" source)
    (handler-case
        (progn
          (unless (and (not (equal source destination))
                       (probe-file destination)
                       (>= (file-write-date destination)
                           (file-write-date source)))
            (convert-media  (content-type record) (first conversion)
                            source destination :args (rest conversion)))
          (setf (pending-status pending) :completed))
      (error(e) (declare (ignore e)) (setf (pending-status pending) :error)))
    (with-lock ((pending-lock server))
      (setf (pending-conversions server)
            (delete pending (pending-conversions server))))))

(defun conversion-loop(server)
  (setf (status server) :running)
  (loop
   (when (eql (status server) :stopping) (return))
   (unless (find :pending (pending-conversions server)
                 :key #'pending-status)
   (with-lock ((pending-lock server))
     (condition-wait (pending-wait server) (pending-lock server))))
   (loop
    (let ((pending (find :pending (pending-conversions server)
                         :key #'pending-status)))
      (unless pending (return))
      (process-pending server pending)
      (condition-broadcast (pending-wait server)))))
  (setf (status server) :stopped))

;; the media variant key for a media server is a list of the media-key and
;; variant position in the media-record
(defmethod media-variant((record media-record) (server media-server)
                         content-type &optional args)
  (when (and (not args) (equal content-type (content-type record)))
    (return-from media-variant
      (values (list (media-key record))
              (probe-file (media-source-path record)))))
  (let* ((c (cons content-type args))
         (p
          (with-lock((lock server))
            (or (conversion-subkey c record)
                (progn
                  (setf (variants record)
                        (nconc (variants record) (list c)))
                  (update-record-from-instance record)
                  (add-pending c record server)
                  (conversion-subkey c record))))))
    (values
     (list (media-key record) p)
     (when (probe-file (variant-path  p record)) t))))

(defmethod media-variant((media-key string) (server media-server)
                            content-type &optional args)
  (assert (get-dictionary media-key (media-records server))
          (media-key)
          "No media ~S found" media-key)
  (media-variant
   (get-dictionary media-key (media-records server))
   server content-type args))

(defmethod media-variant((source pathname) (server media-server)
                            content-type &optional args)
  (media-variant
   (get-dictionary source (media-sources server))
   server
   content-type
   args))

(defun variant-key-to-string(variant-key)
  (format nil "~A-~@[~D~]" (first variant-key) (second variant-key)))

(defun string-to-variant-key(name)
   (cl-ppcre:register-groups-bind  (key (#'parse-integer position))
       ("^(.*)-(\\d+)?$" name)
     (list key position)))

(defmethod media-variant-url(variant-key (server media-server))
  (format nil "~A~A" (media-base-url server)
          (etypecase variant-key
            (string variant-key)
            (list (variant-key-to-string variant-key)))))

(defmethod media-metadata(media-key (server media-server))
  (metadata (get-dictionary media-key (media-records server))))

(defmethod (setf media-metadata)((values list) media-key (server media-server))
  (let ((record  (get-dictionary media-key (media-records server))))
    (prog1
        (setf (metadata record) values)
      (update-record-from-instance record))))

(defmethod media-variant-data((variant-key string) (server media-server))
  (media-variant-data (string-to-variant-key variant-key) server))

(defmethod media-variant-data((variant-key list) (server media-server))
  (multiple-value-bind(path content-type)
      (media-variant-path variant-key server)
    (values (read-media path :content-type content-type) content-type)))

(defmethod media-variant-path((variant-key list) (server media-server))
  (let* ((record (get-dictionary (first variant-key) (media-records server)))
         (subkey (second variant-key)))
    (if subkey
        (let*((variant (elt (variants record) subkey))
              (content-type (first variant))
              (source (media-source-path record))
              (destination (variant-path subkey record)))
          (when (and (not (find-pending variant record server))
                     (probe-file destination)
                     (> (file-write-date destination)
                        (file-write-date source)))
              (return-from media-variant-path
                (values destination content-type)))
          (let ((pending (add-pending variant record server -20)))
            (loop
             (with-lock ((pending-lock server))
               (condition-wait (pending-wait server) (pending-lock server)))
               (case (pending-status pending)
                 (:error  "Unable to complete media conversion ~S" pending)
                 (:completed
                  (if (probe-file destination)
                      (return-from media-variant-path
                        (values destination content-type))
                      (error "Missing media conversion at ~A"
                             destination)))))))
    (values (media-source-path record)  (content-type record)))))

#|
(setq *media-server* (make-instance 'media-server :path #p"/home/willijar/tmp/media/"))
(register-media #p"/home/willijar/teaching/modules/EE401B/3-pulse-modulation/adaptive-prediction-be.eps" *media-server*)

(media-variant "adaptive-prediction-be-0" *media-server* "application/postscript")
(media-variant "adaptive-prediction-be" *media-server* "image/png")
(length (media-variant-data '("adaptive-prediction-be" 0)  *media-server*))

(register-media "$E=mc^2$" *media-server* :content-type "text/x-eqn")


|#
